/* bmpsuite.c
 * Creates a variety of Windows BMP files.
 *
 * By Jason Summers  Apr. 2001
 * <jason1@pobox.com>  http://pobox.com/~jason1/bmpsuite/
 *
 * Can be compiled as a "console mode" Win32 program
 *
 * This program gone through several major rewrites, and
 * is somewhat of a mess at this point.
 *
 * Note that compressed BMP files are currently unsupported.
 *
 */

#include <windows.h>
#include <stdio.h>

#define uchar unsigned char

char *testtext[] =  {
	"00000000000000000",
	"01110111011101110",
	"00100100010000100",
	"00100110011100100",
	"00100100000100100",
	"00100111011100100",
	"00000000000000000" };

#define TEXTWIDTH 17
#define TEXTHEIGHT 7

unsigned char pal_tmp1[] = { 64,128,192,255 };
unsigned char pal_tmp2[] = { 0,85,170, 255 };

#define NORMAL_WIDTH   127
#define NORMAL_HEIGHT  64

/* IMAGE_* is used to select which image you want to
 * create. This is the visual look of the image, not
 * the internal structure */
#define IMAGE_1A  10     // 1bpp, white on black
#define IMAGE_1B  11     // 1bpp, black on white
#define IMAGE_1C  12     // 1bpp, blue and green
#define IMAGE_4A  20     // 4bpp, normal
/* #define IMAGE_4B  21  */   // 4bpp, grayscale only
#define IMAGE_8A  30     // 8bpp, normal
/* #define IMAGE_8B  31  */   // 8bpp, grayscale only
#define IMAGE_24A 40     // 24bpp, normal

#define BITFIELDS_NONE  0
#define BITFIELDS_15    1
#define BITFIELDS_16    2
#define BITFIELDS_32    3

typedef struct makebmpstruct_ {
	int bpp;
	int image;               // what image to create; IMAGE_*
	int compression;         // 0(none) or 1(RLE); set to 0 if using bitfields
	int width, height;
	int version;             // 2=OS/2, 3=V3(normal)
	int pal_entries;         // actual number of entries in palette
	int pal_clrused;         // palette entries to record in header
	int pal_important;       // value of ClrImportant field
	int size_image;          // the biSizeImage field; use -1 to have it calculated
	int xres, yres;          // resolution in pixels per meter
	int bf_format;           // see BITFIELDS_*
	int extra_bytes_after_pal;
	int topdown;

	//----------  for use by makebmp()
	int palentsize;   // 3 or 4 (bytes per palette entry)
	uchar *palettep;

} makebmpstruct;

void setpal(makebmpstruct *b, int n, uchar re, uchar gr, uchar bl)
{
	b->palettep[b->palentsize*n  ] = bl;
	b->palettep[b->palentsize*n+1] = gr;
	b->palettep[b->palentsize*n+2] = re;
}

int makebmp(makebmpstruct *b, char *fn)
{
	int height,width;
	int bytesperrow;
	int bpp;
	BITMAPINFOHEADER bi;
	BITMAPFILEHEADER bf;
	BITMAPCOREHEADER bch;
	DWORD bitfields[3];
	RGBQUAD palrgb[256];
	RGBTRIPLE * palrgbt;   // alias of palrgb;
	int bf_bytes;
	uchar *bits;
	int pal_bytes;   // number of bytes in palette
	int bitssize;    // bytes in image portion
	HANDLE hfile;
	int x,y;
	DWORD written;
	int i;
	unsigned int v;   // temp int
	unsigned char c;  // temp byte
	int os2;
	int header_bytes;
	uchar* temp[1000];

	os2= (b->version==2);

	if(os2) header_bytes=12;
	else    header_bytes=40;

	bpp = b->bpp;
	width = b->width;
	height = b->height;

	bf_bytes = (b->bf_format==BITFIELDS_NONE)?0:12;

	ZeroMemory((void*)&bi,40);     // bitmapinfoheader

	if(os2) ZeroMemory((void*)&bch,12);      // bitmapcoreheader

	ZeroMemory((void*)&bf,14);     // bitmapfileheader 

	if(os2) b->palentsize = 3;
	else    b->palentsize = 4;

	pal_bytes = b->pal_entries * b->palentsize;

	bytesperrow=(((width * bpp)+31)/32)*4;
	bitssize= bytesperrow*height;

	bits=GlobalAlloc(GPTR,bitssize);

	((char*)&bf)[0]='B';
	((char*)&bf)[1]='M';
	bf.bfOffBits= 14+header_bytes+pal_bytes+bf_bytes+b->extra_bytes_after_pal;
	bf.bfSize=    14+header_bytes+pal_bytes+bf_bytes+b->extra_bytes_after_pal+bitssize;

	bi.biSize= header_bytes;
	bi.biWidth= width;
	bi.biHeight= height;
	bi.biPlanes=1;
	bi.biBitCount=bpp;

	if(b->bf_format==BITFIELDS_NONE) {
		if(b->compression) {
			if(b->bpp==4) bi.biCompression=BI_RLE4;
			else if(b->bpp==8) bi.biCompression=BI_RLE8;
			else return 0;
		}
		else {
			bi.biCompression=BI_RGB;
		}
	}
	else {
		bi.biCompression=BI_BITFIELDS;
	}

	if(b->size_image<0) {
		bi.biSizeImage= bitssize;
	}
	else {
		bi.biSizeImage= b->size_image;
	}

	bi.biXPelsPerMeter=   b->xres;
	bi.biYPelsPerMeter=   b->yres;
	bi.biClrUsed=         b->pal_clrused;
	bi.biClrImportant=    b->pal_important;

	if(os2) {
		bch.bcSize=12;
		bch.bcWidth= b->width;
		bch.bcHeight=b->height;
		bch.bcPlanes=1;
		bch.bcBitCount=b->bpp;
	}

	switch(b->bf_format) {
	case BITFIELDS_15:
		bitfields[0]= 0x00007c00;   // red
		bitfields[1]= 0x000003e0;   // green
		bitfields[2]= 0x0000001f;   // blue
		break;
	case BITFIELDS_16:
		bitfields[0]= 0x0000f800;   // red
		bitfields[1]= 0x000007e0;   // green
		bitfields[2]= 0x0000001f;   // blue
		break;
	case BITFIELDS_32:
		bitfields[0]= 0x00ff0000;   // red
		bitfields[1]= 0x0000ff00;   // green
		bitfields[2]= 0x000000ff;   // blue
		break;
	}

	ZeroMemory((void*)palrgb, 256*4);
	palrgbt = (RGBTRIPLE*)palrgb;
	b->palettep = (uchar*)palrgb;

	// create palette
	switch(b->image) {
	case IMAGE_1A:     // 1bpp, black-white
		setpal(b,0,  0,  0,  0);
		setpal(b,1,255,255,255);
		break;
	case IMAGE_1B:   // 1bpp, white-black
		setpal(b,0,255,255,255);
		setpal(b,1,  0,  0,  0);
		break;
	case IMAGE_1C:
		setpal(b,0,64,64,255);
		setpal(b,1,64,255,64);
		break;
	case IMAGE_4A:
		for(i=0;i<4;i++) {
			setpal(b, 0+i,pal_tmp2[i],pal_tmp2[i],pal_tmp2[i]);
			setpal(b, 4+i,pal_tmp1[i],0,0);
			setpal(b, 8+i,0,pal_tmp1[i],0);
			setpal(b,12+i,0,0,pal_tmp1[i]);
		}
		break;

	case IMAGE_8A:

		for(i=0;i<64;i++) {
			c=  (uchar) ((double)i* (255.0/63.0));
			if(i==63) c=255;

			setpal(b,    i,c,c,c);
			setpal(b, 64+i,c,0,0);
			setpal(b,128+i,0,c,0);
			setpal(b,192+i,0,0,c);
		}

		break;

	case IMAGE_24A:
		break;
	default:
		printf("unsupported image type\n"); return 0;
	}

	switch(b->image) {
	case IMAGE_1A:
	case IMAGE_1B:
	case IMAGE_1C:

		for(y=0;y<height;y++) {       // 4 bpp
			for(x=0;x<width;x++) {
				v =  ((x+y)/13)%2;

				if(x<TEXTWIDTH && (height-1-y)<TEXTHEIGHT) {
					if(testtext[height-1-y][x]=='1') {
						v=1;
					}
					else {
						v=0;
					}
				}
				if(b->image==IMAGE_1B) v=1-v;
				if(b->pal_entries==1) { v=0; }
				bits[y*bytesperrow + x/8] |= (v<<(7-x%8));
			}
		}
		break;

	case IMAGE_4A:       // 4 bpp

		for(y=0;y<height;y++) {
			for(x=0;x<width;x++) {

				switch(x/32) {
				case 0: v= 4+   y/16; break;
				case 1: v= 8+(3-y/16); break;
				case 2: v= 12+  y/16; break;
				case 3: v=   (3-y/16); break;
				}

				if(b->pal_entries==4) { v=v%4; }

				if(x<TEXTWIDTH && (height-1-y)<TEXTHEIGHT) {
					if(testtext[height-1-y][x]=='1') {
						v=3;   // palette entry for white
					}
					else {
						v=0;
					}
				}

				if(x%2) {
					bits[y*bytesperrow+x/2] |= v;
				}
				else {
					bits[y*bytesperrow+x/2] |= (v<<4);
				}
			}
		}
		break;

	case IMAGE_8A:

		for(y=0;y<height;y++) {
			for(x=0;x<width;x++) {
				if(x<32) {
					v=64+y;
				}
				else if(x<64) {
					v= 191-y;
				}
				else if(x<96) {
					v=192+y;
				}
				else {
					v= 63-y;
				}

				if(x<TEXTWIDTH && (height-1-y)<TEXTHEIGHT) {
					if(testtext[height-1-y][x]=='1') {
						v=63;   // palette entry for white
					}
					else {
						v=0;
					}
				}
				if(b->pal_entries==64) v=v%64;
				bits[y*bytesperrow+x]=v;
			}
		}
		break;

	case IMAGE_24A:    // truecolor
		for(y=0;y<height;y++) {
			for(x=0;x<width;x++) {
				unsigned int red, green, blue;

				if(x<32) {
					red= 4*y;
					green= (x%32)*8;
					blue= (x%32)*8;
				}
				else if(x<64) {
					red= (x%32)*8;
					green= 4*y;
					blue= (x%32)*8;
				}
				else if(x<96) {
					red= (x%32)*8;
					green= (x%32)*8;
					blue= 4*y;
				}
				else {
					red=green=blue=  y+96;
					blue+= (x%32);
					
				}

				if(x<TEXTWIDTH && (height-1-y)<TEXTHEIGHT) {
					if(testtext[height-1-y][x]=='1') {
						red=green=blue=255;
					}
					else {
						red=green=blue=0;
					}
				}

				if(bpp==16) {
					if(b->bf_format==BITFIELDS_16)
						v = ((red>>3)<<11) | ((green>>2)<<5) | ((blue>>3)); // 565
					else
						v = ((red>>3)<<10) | ((green>>3)<<5) | ((blue>>3)); // 555

					bits[y*bytesperrow + x*2 + 0]= v&0xff;
					bits[y*bytesperrow + x*2 + 1]= v>>8;
				}
				else if(bpp==24) {
					bits[y*bytesperrow + x*3 + 0]= blue;
					bits[y*bytesperrow + x*3 + 1]= green;
					bits[y*bytesperrow + x*3 + 2]= red;
				}
				else if(bpp==32) {        // 32bits
					bits[y*bytesperrow + x*4 + 0]= blue;
					bits[y*bytesperrow + x*4 + 1]= green;
					bits[y*bytesperrow + x*4 + 2]= red;
					bits[y*bytesperrow + x*4 + 3]= 0;
				}
			}
		}
		break;
	default:
		printf("unsupported image type\n"); return 0;
	}

	printf("%s\n",fn);

	hfile=CreateFile(fn,GENERIC_WRITE,FILE_SHARE_READ,NULL,
		CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL,NULL);
	if(hfile==INVALID_HANDLE_VALUE) {
		printf("can't write %s\n",fn);
		GlobalFree(bits);
		return 1;
	}

	WriteFile(hfile,(void*)&bf,14,&written,NULL);
	if(os2) {
		WriteFile(hfile,(void*)&bch,12,&written,NULL);
	}
	else {
		WriteFile(hfile,(void*)&bi,40,&written,NULL);
	}

	if(pal_bytes>0) {
		WriteFile(hfile,(void*)&palrgb,pal_bytes,&written,NULL);
	}

	if(bf_bytes>0) {
		WriteFile(hfile,(void*)&bitfields,bf_bytes,&written,NULL);
	}

	if(b->extra_bytes_after_pal>0) {
		for(i=0;i<b->extra_bytes_after_pal;i++) temp[i]=0;
		WriteFile(hfile,(void*)temp,b->extra_bytes_after_pal,&written,NULL);
	}

	WriteFile(hfile,(void*)bits,bitssize,&written,NULL);
	CloseHandle(hfile);

	GlobalFree(bits);
	return 1;
}

void defaultstruct(makebmpstruct *b)
{
	memset(b,sizeof(makebmpstruct),0);

	b->bpp=24;
	b->image=IMAGE_24A;
	b->compression=0;
	b->width=NORMAL_WIDTH;
	b->height=NORMAL_HEIGHT;
	b->version=3;
	b->pal_entries=0;
	b->pal_clrused=0;
	b->pal_important=0;
	b->size_image= -1;
	b->xres=0;
	b->yres=0;
	b->bf_format=BITFIELDS_NONE;
	b->extra_bytes_after_pal=0;
	b->topdown=0;
}

int main(int argc,char**argv)
{
	makebmpstruct bstruct;
	makebmpstruct *b;

	b=&bstruct;

	defaultstruct(b); b->bpp=1; b->image=IMAGE_1A; b->pal_entries=2;
	makebmp(b,"g01bw.bmp");

	defaultstruct(b); b->bpp=1; b->image=IMAGE_1B; b->pal_entries=2;
	makebmp(b,"g01wb.bmp");

	defaultstruct(b); b->bpp=1; b->image=IMAGE_1C; b->pal_entries=2;
	makebmp(b,"g01bg.bmp");

	defaultstruct(b); b->bpp=1; b->image=IMAGE_1C; b->pal_entries=1;
	b->pal_clrused=1; makebmp(b,"g01p1.bmp");

	defaultstruct(b); b->bpp=4; b->image=IMAGE_4A; b->pal_entries=16;
	makebmp(b,"g04.bmp");

	defaultstruct(b); b->bpp=4; b->image=IMAGE_4A; b->pal_entries=4;
	b->pal_clrused=4; makebmp(b,"g04p4.bmp");

	defaultstruct(b); b->bpp=8; b->image=IMAGE_8A; b->pal_entries=256;
	makebmp(b,"g08.bmp");

	defaultstruct(b); b->bpp=8; b->image=IMAGE_8A; b->pal_entries=256;
	b->extra_bytes_after_pal=100; makebmp(b,"g08offs.bmp");

	defaultstruct(b); b->bpp=8; b->image=IMAGE_8A; b->pal_entries=256;
	b->version=2; makebmp(b,"g08os2.bmp");

	defaultstruct(b); b->bpp=8; b->image=IMAGE_8A; b->pal_entries=256;
	b->width=126; b->height=63; makebmp(b,"g08w126.bmp");

	defaultstruct(b); b->bpp=8; b->image=IMAGE_8A; b->pal_entries=256;
	b->width=125; b->height=62; makebmp(b,"g08w125.bmp");

	defaultstruct(b); b->bpp=8; b->image=IMAGE_8A; b->pal_entries=256;
	b->width=124; b->height=61; makebmp(b,"g08w124.bmp");

	defaultstruct(b); b->bpp=8; b->image=IMAGE_8A; b->pal_entries=256;
	b->pal_clrused=256; makebmp(b,"g08p256.bmp");

	defaultstruct(b); b->bpp=8; b->image=IMAGE_8A; b->pal_entries=256;
	b->pal_clrused=256; b->pal_important=256; makebmp(b,"g08pi256.bmp");

	defaultstruct(b); b->bpp=8; b->image=IMAGE_8A; b->pal_entries=256;
	b->pal_clrused=256; b->pal_important=64; makebmp(b,"g08pi64.bmp");

	defaultstruct(b); b->bpp=8; b->image=IMAGE_8A; b->pal_entries=256;
	b->xres=7874; b->yres=7874; makebmp(b,"g08res22.bmp");

	defaultstruct(b); b->bpp=8; b->image=IMAGE_8A; b->pal_entries=256;
	b->xres=7874; b->yres=3937; makebmp(b,"g08res21.bmp");

	defaultstruct(b); b->bpp=8; b->image=IMAGE_8A; b->pal_entries=256;
	b->xres=3937; b->yres=3937; makebmp(b,"g08res11.bmp");

	defaultstruct(b); b->bpp=8; b->image=IMAGE_8A; b->pal_entries=64;
	b->pal_clrused=64; makebmp(b,"g08p64.bmp");

	defaultstruct(b); b->bpp=8; b->image=IMAGE_8A; b->pal_entries=256;
	b->size_image=0; makebmp(b,"g08s0.bmp");

	defaultstruct(b); b->bpp=16; b->image=IMAGE_24A;
	makebmp(b,"g16def555.bmp");

	defaultstruct(b); b->bpp=16; b->image=IMAGE_24A;
	b->bf_format=BITFIELDS_15; makebmp(b,"g16bf555.bmp");

	defaultstruct(b); b->bpp=16; b->image=IMAGE_24A;
	b->bf_format=BITFIELDS_16; makebmp(b,"g16bf565.bmp");

	defaultstruct(b); b->bpp=24; b->image=IMAGE_24A;
	makebmp(b,"g24.bmp");

	defaultstruct(b); b->bpp=32; b->image=IMAGE_24A;
	makebmp(b,"g32def.bmp");

	defaultstruct(b); b->bpp=32; b->image=IMAGE_24A;
	b->bf_format=BITFIELDS_32; makebmp(b,"g32bf.bmp");

	return 0;
}
